import time
import argparse
from keras.callbacks import ModelCheckpoint
from keras.models import Sequential, Model
from keras.layers import Dense, Dropout,Input,GlobalMaxPooling1D
from keras.layers import Embedding
from keras.layers import Conv1D, GlobalAveragePooling1D, MaxPooling1D
import numpy as np
import pandas as pd
from utils import get_data_splits, loadEmbeddingMatrix, vectorize, load_data, load_embeddings, preprocess_text
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score, accuracy_score
from sklearn.preprocessing import LabelEncoder
from keras.initializers import Constant
from keras.layers import Input, Embedding, Dense, Conv2D, MaxPool2D
from keras.layers import Reshape, Flatten, Concatenate, Dropout, SpatialDropout1D
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences



def get_model(sentence_maxlen, embedding_matrix, embd_dim, word_index,s_dropout=0.4, dropout=0.1):
	filter_sizes = [1,2,3,5]
	num_filters = 36
	inp = Input(shape=(sentence_maxlen,))
	x = Embedding(len(word_index), embd_dim, weights=[embedding_matrix])(inp)
	x = SpatialDropout1D(s_dropout)(x)
	x = Reshape((sentence_maxlen, embd_dim, 1))(x)
	maxpool_pool = []
	for i in range(len(filter_sizes)):
		conv = Conv2D(num_filters, kernel_size=(filter_sizes[i], embd_dim),
									 kernel_initializer='he_normal', activation='elu')(x)
		maxpool_pool.append(MaxPool2D(pool_size=(sentence_maxlen - filter_sizes[i] + 1, 1))(conv))
	z = Concatenate(axis=1)(maxpool_pool)   
	z = Flatten()(z)
	z = Dropout(dropout)(z)
	outp = Dense(1, activation="sigmoid")(z)
	model = Model(inputs=inp, outputs=outp)
	model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
	return model

def downSampling(idx_list, labels, rr=1.):
	np_num = len([l for l in labels if l==0])
	p_num = len(labels) - np_num
	ratio = p_num*rr/np_num
	new_idx = [ii  for ii in range(idx_list.shape[0]) if labels[ii]==1 or (labels[ii]==0 and np.random.random()<ratio)]
	return idx_list[(new_idx)], labels[(new_idx)]

def run_model(train,test, embd):
	np.random.seed(42)
	down_sample = True
	# cut texts after this number of words
	# (among top max_features most common words)
	#GLOVE_DIR = "./resources/glove"
	#embd_dim = 300
	train_data = load_data(train)
	test_data = load_data(test)

	data = train_data+test_data

	#train, dev, test = get_data_splits(df)
	#train_data = load_data(train)
	#dev_data = load_data(dev)
	#test_data = load_data(test)
	sentence_maxlen = max(map(len, (d for d, _ in data)))

	vocab = []
	for d, _ in train_data+test_data:
		for w in d:
			if w not in vocab: vocab.append(w)
	vocab = sorted(vocab)
	print('vocab size', len(vocab))
	w2i = {w:i for i,w in enumerate(vocab)}

	np.random.shuffle(train_data)

	X_train, Y_train = vectorize(train_data, sentence_maxlen, w2i)
	X_test, Y_test = vectorize(test_data, sentence_maxlen, w2i)
	
	if down_sample:
		X_train, Y_train = downSampling(X_train, Y_train)
	print(X_train.shape)
	print(X_test.shape)

	embd_matrix, embd_dim = loadEmbeddingMatrix(embd, w2i) #load_glove_weights('./resources/glove', embd_dim, vocab_size, w2i)
	epochs = 5
	batch_size = 64

	filepath="./models_chkpt/" + embd + "/cnn_best_{epoch:02d}-{val_acc:.2f}.hdf5"
	checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
	callbacks_list = [checkpoint]
 
	model = get_model(sentence_maxlen, embd_matrix, embd_dim, w2i)
	print(model.summary())
	print("Traning Model...")

	model.fit(X_train, Y_train,batch_size=32,epochs=epochs,callbacks=callbacks_list, validation_data=(X_test, Y_test))
	Y_pred = model.predict(X_test)
	return Y_pred #f1_score(Y_test, Y_pred), accuracy_score(Y_test, Y_pred)

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description="Sentence classification for propaganda detection")
	parser.add_argument('--fold', help="The fold to use", default=0)
	parser.add_argument('--embd', help="Embeddings", default="glove")
	parser.add_argument('--epochs', default=5, type=int)
	parser.add_argument('--downsample', default=0)
	parser.add_argument("--batch_size", default=32, type=int)
	parser.add_argument("--dropout", default=0.1, type=float)
	parser.add_argument("--s_dropout", default=0.4, type=float)
	parser.add_argument("--preprocess", default=0)
	args = parser.parse_args()

	f1_prop = [] 
	f1_nprop = []
	accs = []
	trained= False

	if args.fold == "all":
		for fold in [0,1,2]:
			print('fold:', fold)
			train,test = get_data_splits(fold)
			le = LabelEncoder().fit(train['label'])
			train['label'] = le.transform(train['label'])
			test['label'] = le.transform(test['label'])
			y_true = test['label']

			if args.preprocess:
				train = preprocess_text(train)
				train = train[train['sentence_text'] != '']
				test = preprocess_text(test)
				test = test[test['sentence_text'] != '']

			np.random.seed(42)

			train_data = load_data(train)
			test_data = load_data(test)



			data = train_data+test_data

			sentence_maxlen = max(map(len, (d for d, _ in data)))

			vocab = []
			for d, _ in data:
				for w in d:
					if w not in vocab: vocab.append(w)
			vocab = sorted(vocab)
			w2i = {w:i for i,w in enumerate(vocab)}

			np.random.shuffle(train_data)

			X_train, Y_train = vectorize(train_data, sentence_maxlen, w2i)
			X_test, Y_test = vectorize(test_data, sentence_maxlen, w2i)

			X_train, X_dev, Y_train, Y_dev = train_test_split(X_train, Y_train, test_size=0.1, stratify=Y_train)
			
			if args.downsample:
				X_train, Y_train = downSampling(X_train, Y_train)
			if not trained:
				embd_matrix, embd_dim = loadEmbeddingMatrix(args.embd, w2i) #load_glove_weights('./resources/glove', embd_dim, vocab_size, w2i)
				trained=True
			#embd_matrix, embd_dim =np.zeros((len(w2i), 300)), 300
			strtime = str(time.time())
			weights_fp = "./models_chkpt/" + strtime + "_weights.hdf5"
			model_fp = "./models_chkpt/" + strtime + "_model.json"
			checkpoint = ModelCheckpoint(weights_fp, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
			callbacks_list = [checkpoint]
			model = get_model(sentence_maxlen, embd_matrix, embd_dim, w2i, dropout=args.dropout, s_dropout=args.s_dropout)

			model.fit(X_train, Y_train,batch_size=args.batch_size,epochs=args.epochs,callbacks=callbacks_list, validation_data=(X_dev, Y_dev))
			# load best weights and serialize
			model.load_weights(filepath)
			model.save_weights(weights_fp)
			with open(model_fp, 'w') as f:
				f.write(model.to_json())
			# prediction
			Y_pred = model.predict(X_test)
			Y_pred = np.round(Y_pred).astype(int)
			f1_p = f1_score(Y_test, Y_pred, pos_label=1)
			f1_np = f1_score(Y_test, Y_pred, pos_label=0)
			acc = accuracy_score(Y_test, Y_pred)
			f1_prop.append(f1_p)
			f1_nprop.append(f1_np)
			accs.append(acc)
			print("fold", fold, "f1_p:", f1_p, "f1_np:",f1_np, "acc:", acc)
	else:
		print('fold:', args.fold)
		train,test = get_data_splits(args.fold)
		le = LabelEncoder().fit(train['label'])
		train['label'] = le.transform(train['label'])
		test['label'] = le.transform(test['label'])
		y_true = test['label']

		np.random.seed(42)

		train_data = load_data(train)
		test_data = load_data(test)

		if args.preprocess:
			train = preprocess_text(train)
			train = train[train['sentence_text'] != '']
			test = preprocess_text(test)
			test = test[test['sentence_text'] != '']

		data = train_data+test_data

		sentence_maxlen = max(map(len, (d for d, _ in data)))

		vocab = []
		for d, _ in data:
			for w in d:
				if w not in vocab: vocab.append(w)
		vocab = sorted(vocab)
		w2i = {w:i for i,w in enumerate(vocab)}

		np.random.shuffle(train_data)

		X_train, Y_train = vectorize(train_data, sentence_maxlen, w2i)
		X_test, Y_test = vectorize(test_data, sentence_maxlen, w2i)

		X_train, X_dev, Y_train, Y_dev = train_test_split(X_train, Y_train, test_size=0.1, stratify=Y_train)
		
		if args.downsample:
			X_train, Y_train = downSampling(X_train, Y_train)

		embd_matrix, embd_dim = loadEmbeddingMatrix(args.embd, w2i) #load_glove_weights('./resources/glove', embd_dim, vocab_size, w2i)
		#embd_matrix, embd_dim =np.zeros((len(w2i), 300)), 300
		
		strtime = str(time.time())
		#filepath="./models_chkpt/" + args.embd + "_cnn_best_{epoch:02d}-{val_acc:.2f}.hdf5"
		filepath = "./models_chkpt/" + strtime + "_weights.hdf5"
		model_fp = "./models_chkpt/" + strtime + "_model.json"
		checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
		#checkpoint2 = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
		callbacks_list = [checkpoint]
	 
		model = get_model(sentence_maxlen, embd_matrix, embd_dim, w2i, dropout=args.dropout, s_dropout=args.s_dropout)
		print("Traning Model...")
		model.fit(X_train, Y_train,batch_size=args.batch_size,epochs=args.epochs,callbacks=callbacks_list, validation_data=(X_dev, Y_dev))
		model.load_weights(filepath)
		model.save_weights(weights_fp)
		with open(model_fp, 'w') as f:
			f.write(model.to_json())
		Y_pred = model.predict(X_test)
		Y_pred = np.round(Y_pred).astype(int)
		f1_p = f1_score(Y_test, Y_pred, pos_label=1)
		f1_np = f1_score(Y_test, Y_pred, pos_label=0)
		acc = accuracy_score(Y_test, Y_pred)
		f1_prop.append(f1_p)
		f1_nprop.append(f1_np)
		accs.append(acc)
		print("fold", args.fold, "f1_p:", f1_p, "f1_np:",f1_np, "acc:", acc)
	print('Finished')
	print("averaged results:", "f1_p:", np.mean(f1_prop), "f1_np:",np.mean(f1_nprop), "acc:", np.mean(accs))
